@@ -1,3 +1,4 @@
+import math
 import re
 import json
 import logging
@@ -20,7 +21,8 @@ from .errors import ValidationError, RequestError
 from .crypto import (
     a32_to_base64, encrypt_key, base64_url_encode, encrypt_attr, base64_to_a32,
     base64_url_decode, decrypt_attr, a32_to_str, get_chunks, str_to_a32,
-    decrypt_key, mpi_to_int, stringhash, prepare_key, make_id, makebyte
+    decrypt_key, mpi_to_int, stringhash, prepare_key, make_id, makebyte,
+    modular_inverse
 )
 
 logger = logging.getLogger(__name__)
@@ -119,29 +121,46 @@ class Mega:
             )
 
             private_key = a32_to_str(rsa_private_key)
-            self.rsa_private_key = [0, 0, 0, 0]
+            # The private_key contains 4 MPI integers concatenated together.
+            rsa_private_key = [0, 0, 0, 0]
             for i in range(4):
-                l = int(
-                    ((private_key[0]) * 256 + (private_key[1]) + 7) / 8
-                ) + 2
-                self.rsa_private_key[i] = mpi_to_int(private_key[:l])
-                private_key = private_key[l:]
+                # An MPI integer has a 2-byte header which describes the number
+                # of bits in the integer.
+                bitlength = (private_key[0] * 256) + private_key[1]
+                bytelength = math.ceil(bitlength / 8)
+                # Add 2 bytes to accommodate the MPI header
+                bytelength += 2
+                rsa_private_key[i] = mpi_to_int(private_key[:bytelength])
+                private_key = private_key[bytelength:]
+
+            first_factor_p = rsa_private_key[0]
+            second_factor_q = rsa_private_key[1]
+            private_exponent_d = rsa_private_key[2]
+            # In MEGA's webclient javascript, they assign [3] to a variable
+            # called u, but I do not see how it corresponds to pycryptodome's
+            # RSA.construct and it does not seem to be necessary.
+            rsa_modulus_n = first_factor_p * second_factor_q
+            phi = (first_factor_p - 1) * (second_factor_q - 1)
+            public_exponent_e = modular_inverse(private_exponent_d, phi)
+
+            rsa_components = (
+                rsa_modulus_n,
+                public_exponent_e,
+                private_exponent_d,
+                first_factor_p,
+                second_factor_q,
+            )
+            rsa_decrypter = RSA.construct(rsa_components)
 
             encrypted_sid = mpi_to_int(base64_url_decode(resp['csid']))
-            rsa_decrypter = RSA.construct(
-                (
-                    self.rsa_private_key[0] * self.rsa_private_key[1], 0,
-                    self.rsa_private_key[2], self.rsa_private_key[0],
-                    self.rsa_private_key[1]
-                )
-            )
-            sid = '%x' % rsa_decrypter.key._decrypt(encrypted_sid)
+
+            sid = '%x' % rsa_decrypter._decrypt(encrypted_sid)
             sid = binascii.unhexlify('0' + sid if len(sid) % 2 else sid)
             self.sid = base64_url_encode(sid[:43])
 
     @retry(
         retry=retry_if_exception_type(RuntimeError),
-        wait=wait_exponential(multiplier=1, min=1, max=20)
+        wait=wait_exponential(multiplier=2, min=2, max=60)
     )
     def _api_request(self, data):
         params = {'id': self.sequence_num}
@@ -284,9 +303,8 @@ class Mega:
             if foldername != '':
                 for file in files.items():
                     if (
-                        file[1]['a'] and
-                        file[1]['t'] and
-                        file[1]['a']['n'] == foldername
+                        file[1]['a'] and file[1]['t']
+                        and file[1]['a']['n'] == foldername
                     ):
                         if parent_desc == file[1]['p']:
                             parent_desc = file[0]
@@ -314,23 +332,20 @@ class Mega:
                     parent_dir_name, files=files
                 )
                 if (
-                    filename and parent_node_id and
-                    file[1]['a'] and file[1]['a']['n'] == filename and
-                    parent_node_id == file[1]['p']
+                    filename and parent_node_id and file[1]['a']
+                    and file[1]['a']['n'] == filename
+                    and parent_node_id == file[1]['p']
                 ):
                     if (
-                        exclude_deleted and
-                        self._trash_folder_node_id == file[1]['p']
+                        exclude_deleted
+                        and self._trash_folder_node_id == file[1]['p']
                     ):
                         continue
                     return file
-            if (
-                filename and
-                file[1]['a'] and file[1]['a']['n'] == filename
-            ):
+            if (filename and file[1]['a'] and file[1]['a']['n'] == filename):
                 if (
-                    exclude_deleted and
-                    self._trash_folder_node_id == file[1]['p']
+                    exclude_deleted
+                    and self._trash_folder_node_id == file[1]['p']
                 ):
                     continue
                 return file
@@ -568,7 +583,7 @@ class Mega:
         """
         Download a file by it's file object
         """
-        self._download_file(
+        return self._download_file(
             file_handle=None,
             file_key=None,
             file=file[1],
@@ -579,13 +594,13 @@ class Mega:
 
     def _export_file(self, node):
         node_data = self._node_data(node)
-        self._api_request([
-            {
+        self._api_request(
+            [{
                 'a': 'l',
                 'n': node_data['h'],
                 'i': self.request_id
-            }
-        ])
+            }]
+        )
         return self.get_link(node)
 
     def export(self, path=None, node_id=None):
@@ -608,7 +623,9 @@ class Mega:
 
         master_key_cipher = AES.new(a32_to_str(self.master_key), AES.MODE_ECB)
         ha = base64_url_encode(
-            master_key_cipher.encrypt(node_data['h'] + node_data['h'])
+            master_key_cipher.encrypt(
+                node_data['h'].encode("utf8") + node_data['h'].encode("utf8")
+            )
         )
 
         share_key = secrets.token_bytes(16)
@@ -646,7 +663,7 @@ class Mega:
         path = self._parse_url(url).split('!')
         file_id = path[0]
         file_key = path[1]
-        self._download_file(
+        return self._download_file(
             file_handle=file_id,
             file_key=file_key,
             dest_path=dest_path,
@@ -694,7 +711,7 @@ class Mega:
             iv = file['iv']
             meta_mac = file['meta_mac']
 
-        # Seems to happens sometime... When  this occurs, files are
+        # Seems to happens sometime... When this occurs, files are
         # inaccessible also in the official also in the official web app.
         # Strangely, files can come back later.
         if 'g' not in file_data:
@@ -716,51 +733,55 @@ class Mega:
         else:
             dest_path += '/'
 
-        temp_output_file = tempfile.NamedTemporaryFile(
+        with tempfile.NamedTemporaryFile(
             mode='w+b', prefix='megapy_', delete=False
-        )
+        ) as temp_output_file:
+            k_str = a32_to_str(k)
+            counter = Counter.new(
+                128, initial_value=((iv[0] << 32) + iv[1]) << 64
+            )
+            aes = AES.new(k_str, AES.MODE_CTR, counter=counter)
 
-        k_str = a32_to_str(k)
-        counter = Counter.new(128, initial_value=((iv[0] << 32) + iv[1]) << 64)
-        aes = AES.new(k_str, AES.MODE_CTR, counter=counter)
+            mac_str = '\0' * 16
+            mac_encryptor = AES.new(
+                k_str, AES.MODE_CBC, mac_str.encode("utf8")
+            )
+            iv_str = a32_to_str([iv[0], iv[1], iv[0], iv[1]])
 
-        mac_str = '\0' * 16
-        mac_encryptor = AES.new(k_str, AES.MODE_CBC, mac_str)
-        iv_str = a32_to_str([iv[0], iv[1], iv[0], iv[1]])
+            for chunk_start, chunk_size in get_chunks(file_size):
+                chunk = input_file.read(chunk_size)
+                chunk = aes.decrypt(chunk)
+                temp_output_file.write(chunk)
 
-        for chunk_start, chunk_size in get_chunks(file_size):
-            chunk = input_file.read(chunk_size)
-            chunk = aes.decrypt(chunk)
-            temp_output_file.write(chunk)
+                encryptor = AES.new(k_str, AES.MODE_CBC, iv_str)
+                for i in range(0, len(chunk) - 16, 16):
+                    block = chunk[i:i + 16]
+                    encryptor.encrypt(block)
+
+                # fix for files under 16 bytes failing
+                if file_size > 16:
+                    i += 16
+                else:
+                    i = 0
 
-            encryptor = AES.new(k_str, AES.MODE_CBC, iv_str)
-            for i in range(0, len(chunk) - 16, 16):
                 block = chunk[i:i + 16]
-                encryptor.encrypt(block)
+                if len(block) % 16:
+                    block += b'\0' * (16 - (len(block) % 16))
+                mac_str = mac_encryptor.encrypt(encryptor.encrypt(block))
 
-            # fix for files under 16 bytes failing
-            if file_size > 16:
-                i += 16
-            else:
-                i = 0
-
-            block = chunk[i:i + 16]
-            if len(block) % 16:
-                block += '\0' * (16 - (len(block) % 16))
-            mac_str = mac_encryptor.encrypt(encryptor.encrypt(block))
-
-            file_info = os.stat(temp_output_file.name)
-            logger.info('%s of %s downloaded', file_info.st_size, file_size)
-
-        file_mac = str_to_a32(mac_str)
-
-        temp_output_file.close()
-
-        # check mac integrity
-        if (file_mac[0] ^ file_mac[1], file_mac[2] ^ file_mac[3]) != meta_mac:
-            raise ValueError('Mismatched mac')
-
-        shutil.move(temp_output_file.name, dest_path + file_name)
+                file_info = os.stat(temp_output_file.name)
+                logger.info(
+                    '%s of %s downloaded', file_info.st_size, file_size
+                )
+            file_mac = str_to_a32(mac_str)
+            # check mac integrity
+            if (
+                file_mac[0] ^ file_mac[1], file_mac[2] ^ file_mac[3]
+            ) != meta_mac:
+                raise ValueError('Mismatched mac')
+            output_path = Path(dest_path + file_name)
+            shutil.move(temp_output_file.name, output_path)
+            return output_path
 
     def upload(self, filename, dest=None, dest_filename=None):
         # determine storage node
@@ -787,7 +808,9 @@ class Mega:
             completion_file_handle = None
 
             mac_str = '\0' * 16
-            mac_encryptor = AES.new(k_str, AES.MODE_CBC, mac_str)
+            mac_encryptor = AES.new(
+                k_str, AES.MODE_CBC, mac_str.encode("utf8")
+            )
             iv_str = a32_to_str([ul_key[4], ul_key[5], ul_key[4], ul_key[5]])
             if file_size > 0:
                 for chunk_start, chunk_size in get_chunks(file_size):
